@@ -1,5 +1,8 @@
 # https://www.cnblogs.com/superbaby11/p/15560994.html
 from threading import Thread
+import io
+import os
+from contextlib import contextmanager
 
 
 def run_cmd(cmd):
@@ -322,6 +325,7 @@ def full_page_snapshot(url, dir_snapshot):
     if "http_proxy" in os.environ:
         del os.environ["http_proxy"]
     chrome_options = webdriver.ChromeOptions()
+    chrome_options.add_experimental_option("excludeSwitches", ["enable-automation"])
     driver = webdriver.Remote(
         command_executor="http://127.0.0.1:4444/wd/hub", options=chrome_options
     )
@@ -381,6 +385,7 @@ def full_page_snapshot(url, dir_snapshot):
         print("main_scroll_height is %d" % main_scroll_height)
 
     driver.execute_script("window.scrollTo(0, 0);")
+    time.sleep(1)
     page_index = 1
     driver.save_screenshot(dir_snapshot + os.path.sep + ("page_%06d.png" % page_index))
 
@@ -401,15 +406,23 @@ def full_page_snapshot(url, dir_snapshot):
         % (page_index, window_page_y_offset, main_scroll_height, page_height)
     )
 
-    while (window_page_y_offset + page_height < main_scroll_height) and (
+    while (int(window_page_y_offset) + int(page_height) < int(main_scroll_height)) and (
         page_index < 100
     ):
         driver.execute_script("window.scrollBy(0, %d);" % page_height)
+        time.sleep(1)
+        last_window_page_y_offset = window_page_y_offset
+        window_page_y_offset = driver.execute_script("return window.pageYOffset")
+        if int(last_window_page_y_offset) == int(window_page_y_offset):
+            print(
+                "last_window_page_y_offset == window_page_y_offset == %d"
+                % window_page_y_offset,
+            )
+            break
         page_index = page_index + 1
         driver.save_screenshot(
             dir_snapshot + os.path.sep + ("page_%06d.png" % page_index)
         )
-        window_page_y_offset = driver.execute_script("return window.pageYOffset")
         print(
             "page_index is %d window_page_y_offset is %d main_scroll_height is %d page_height is %d"
             % (page_index, window_page_y_offset, main_scroll_height, page_height)
@@ -435,12 +448,17 @@ def find_line_in_file(file_path, line_content, search_line_number_begin=0):
 
 
 def dev_checkout_file(filename):
-    cmd_ret = run_cmd("git log -1 %s" % filename)
+    cmd_ret = run_cmd("git log -1 -- %s" % filename)
     if cmd_ret == "":
         import os
 
         if os.path.isfile(filename):
             os.remove(filename)
+        else:
+            parent_dir = os.path.dirname(filename)
+            if parent_dir.strip().__len__() > 0:
+                if not os.path.isdir(parent_dir):
+                    os.makedirs(parent_dir)
     else:
         print_run_cmd("git checkout %s" % filename)
 
@@ -513,3 +531,130 @@ def dev_clean_ipynb():
             move_single_ipynb(i)
         else:
             print("Ignore %s" % i)
+
+
+def empty_file(file_path):
+    import fileinput
+
+    with fileinput.FileInput(file_path, inplace=True) as f:
+        for line in f:
+            # print(line, end="")
+            pass
+
+
+def append_text_file_with_utf8(file_name, file_content):
+    with open(file_name, "a", encoding="utf-8") as file:
+        file.write(file_content)
+
+
+# https://www.zopatista.com/python/2013/11/26/inplace-file-rewriting/
+@contextmanager
+def inplace(
+    filename,
+    mode="r",
+    buffering=-1,
+    encoding=None,
+    errors=None,
+    newline=None,
+    backup_extension=None,
+):
+    """Allow for a file to be replaced with new content.
+
+    yields a tuple of (readable, writable) file objects, where writable
+    replaces readable.
+
+    If an exception occurs, the old file is restored, removing the
+    written data.
+
+    mode should *not* use 'w', 'a' or '+'; only read-only-modes are supported.
+
+    """
+
+    # move existing file to backup, create new file with same permissions
+    # borrowed extensively from the fileinput module
+    if set(mode).intersection("wa+"):
+        raise ValueError("Only read-only file modes can be used")
+
+    backupfilename = filename + (backup_extension or os.extsep + "bak")
+    try:
+        os.unlink(backupfilename)
+    except os.error:
+        pass
+    os.rename(filename, backupfilename)
+    readable = io.open(
+        backupfilename,
+        mode,
+        buffering=buffering,
+        encoding=encoding,
+        errors=errors,
+        newline=newline,
+    )
+    try:
+        perm = os.fstat(readable.fileno()).st_mode
+    except OSError:
+        writable = open(
+            filename,
+            "w" + mode.replace("r", ""),
+            buffering=buffering,
+            encoding=encoding,
+            errors=errors,
+            newline=newline,
+        )
+    else:
+        os_mode = os.O_CREAT | os.O_WRONLY | os.O_TRUNC
+        if hasattr(os, "O_BINARY"):
+            os_mode |= os.O_BINARY
+        fd = os.open(filename, os_mode, perm)
+        writable = io.open(
+            fd,
+            "w" + mode.replace("r", ""),
+            buffering=buffering,
+            encoding=encoding,
+            errors=errors,
+            newline=newline,
+        )
+        try:
+            if hasattr(os, "chmod"):
+                os.chmod(filename, perm)
+        except OSError:
+            pass
+    try:
+        yield readable, writable
+    except Exception:
+        # move backup back
+        try:
+            os.unlink(filename)
+        except os.error:
+            pass
+        os.rename(backupfilename, filename)
+        raise
+    finally:
+        readable.close()
+        writable.close()
+        try:
+            os.unlink(backupfilename)
+        except os.error:
+            pass
+
+
+def update_line_in_file_with_utf8(
+    file_path, old_line_number, new_line_content, flag_replace=True
+):
+    with inplace(file_path, "r", newline="", encoding="utf-8") as (infh, outfh):
+        line_number = 0
+        for line in infh:
+            line_number = line_number + 1
+            if line_number == old_line_number:
+                outfh.write(new_line_content)
+                if not flag_replace:
+                    outfh.write(line)
+            else:
+                outfh.write(line)
+
+
+def replace_line_in_file_with_utf8(file_path, old_line_number, new_line_content):
+    update_line_in_file_with_utf8(file_path, old_line_number, new_line_content, True)
+
+
+def add_line_in_file_with_utf8(file_path, old_line_number, new_line_content):
+    update_line_in_file_with_utf8(file_path, old_line_number, new_line_content, False)
